def test_ocba():
    """
    Test OCBA.

    """

import copy
import numpy as np
from math import inf
from spotPython.fun.objectivefunctions import analytical
from spotPython.spot import spot
from scipy.optimize import shgo
from scipy.optimize import direct
from scipy.optimize import differential_evolution
import matplotlib.pyplot as plt
from numpy import float64
from spotPython.budget.ocba import get_ocba

# Test based on the example from the book:
# Chun-Hung Chen and Loo Hay Lee:
#     Stochastic Simulation Optimization: An Optimal Computer Budget Allocation,
#     pp. 49 and pp. 215
#     p. 49:
#     mean_y = np.array([1,2,3,4,5])
#     var_y = np.array([1,1,9,9,4])
#     get_ocba(mean_y, var_y, 50)
#     [11  9 19  9  2]

fun = analytical().fun_linear
fun_control = {"sigma": 0.001,
               "seed": 123}
spot_1_noisy = spot.Spot(fun=fun,
                   lower = np.array([-1]),
                   upper = np.array([1]),
                   fun_evals = 20,
                   fun_repeats = 2,
                   noise = True,
                   ocba_delta=1,
                   seed=123,
                   show_models=False,
                   fun_control = fun_control,
                   design_control={"init_size": 3,
                                   "repeats": 2},
                   surrogate_control={"noise": True})
spot_1_noisy.run()
spot_2 = copy.deepcopy(spot_1_noisy)
spot_2.mean_y = np.array([1,2,3,4,5])
spot_2.var_y = np.array([1,1,9,9,4])
n = 50
o = get_ocba(spot_2.mean_y, spot_2.var_y, n)
assert sum(o) == 50
assert (o == np.array([[11, 9, 19, 9, 2]])).all()


import copy
from math import erf
import matplotlib.pyplot as plt
from numpy import min, std, var, mean
from numpy import sqrt
from numpy import exp
from numpy import array
from numpy import log
from numpy import power
from numpy import abs
from numpy import eye
from numpy import sum
from numpy import diag
from numpy import pi
from numpy import ones, zeros
from numpy import spacing, empty_like
from numpy import float64
from numpy import append, ndarray, multiply, isinf, linspace, meshgrid, ravel
from numpy.linalg import cholesky, solve, LinAlgError, cond
from scipy.optimize import differential_evolution
from scipy.linalg import cholesky as scipy_cholesky
import pylab
from spotPython.design.spacefilling import spacefilling
from spotPython.build.surrogates import surrogates
from scipy.spatial.distance import squareform
from scipy.spatial.distance import pdist
from scipy.spatial.distance import cdist
from spotPython.utils.repair import repair_non_numeric
from spotPython.utils.aggregate import aggregate_mean_var
import logging

logger = logging.getLogger(__name__)
# configure the handler and formatter as needed
py_handler = logging.FileHandler(f"{__name__}.log", mode='w')
py_formatter = logging.Formatter("%(name)s %(asctime)s %(levelname)s %(message)s")
# add formatter to the handler
py_handler.setFormatter(py_formatter)
# add handler to the logger
logger.addHandler(py_handler)


class Kriging(surrogates):
    def __init__(
            self,
            noise=False,
            cod_type="norm",
            var_type=["num"],
            use_cod_y=False,
            name="kriging",
            seed=124,
            model_optimizer=None,
            model_fun_evals=None,
            min_theta=-3,  # TODO
            max_theta=2,  # TODO
            n_theta=1,
            n_p=1,
            optim_p=False,
            log_level=50,
            **kwargs
    ):
        """
        Kriging surrogate.

        Args:
            noise (bool):
                use regression instead of interpolation kriging. Defaults to "False".
            cod_type (bool):
                normalize or standardize X and values. Can be None, "norm", or "std". Defaults to "norm".
            var_type (str):
                variable type. Can be either `"num`" (numerical) of `"factor"` (factor).
                Defaults to `"num"`.
            use_cod_y (bool):
                use coded y values (instead of natural one). Defaults to `False`.
            name (str):
                Surrogate name. Defaults to `"kriging"`.
            seed (int):
                Random seed. Defaults to `124`.
            model_optimizer (object):
                Optimizer on the surrogate. If `None`, `differential_evolution` is selected.
            model_fun_evals (int):
                Number of iterations used by the optimizer on the surrogate.
            min_theta (float):
                min log10 theta value. Defaults to `-6.`.
            max_theta (float):
                max log10 theta value. Defaults to `3.`.
            n_theta (int):
                number of theta values. Defaults to `1`.
            n_p (int):
                number of p values. Defaults to `1`.
            optim_p (bool):
                Determines whether `p` should be optimized.
            log_level (int):
                logging level, e.g., `20` is `"INFO"`. Defaults to `50` (`"CRITICAL"`).

        Attributes:
            nat_range_X (list):
                List of X natural ranges.
            nat_range_y (list):
                List of y nat ranges.
            noise (bool):
                noisy objective function. Default: False. If `True`, regression kriging will be used.
            var_type (str):
                variable type. Can be either `"num`" (numerical) of `"factor"` (factor).
            num_mask (array):
                array of bool variables. `True` represent numerical (float) variables.
            factor_mask (array):
                array of factor variables. `True` represents factor (unordered) variables.
            int_mask (array):
                array of integer variables. `True` represents integers (ordered) variables.
            ordered_mask (array):
                array of ordered variables. `True` represents integers or float (ordered) variables.
                Set of veriables which an order relation, i.e., they are either num (float) or int.
            name (str):
                Surrogate name
            seed (int):
                Random seed.
            use_cod_y (bool):
                Use coded y values.
            sigma (float):
                Kriging sigma.
            gen (method):
                Design generator, e.g., spotPython.design.spacefilling.spacefilling.
            min_theta (float):
                min log10 theta value. Defaults: -6.
            max_theta (float):
                max log10 theta value. Defaults: 3.
            min_p (float):
                min p value. Default: 1.
            max_p (float):
                max p value. Default: 2.

        Examples:
            Surrogate of the x*sin(x) function.
            See:
            [scikit-learn](https://scikit-learn.org/stable/auto_examples/gaussian_process/plot_gpr_noisy_targets.html)

            >>> from spotPython.build.kriging import Kriging
            >>> import numpy as np
            >>> import matplotlib.pyplot as plt
            >>> rng = np.random.RandomState(1)
            >>> X = linspace(start=0, stop=10, num=1_000).reshape(-1, 1)
            >>> y = np.squeeze(X * np.sin(X))
            >>> training_indices = rng.choice(arange(y.size), size=6, replace=False)
            >>> X_train, y_train = X[training_indices], y[training_indices]
            >>> S = Kriging(name='kriging', seed=124)
            >>> S.fit(X_train, y_train)
            >>> mean_prediction, std_prediction = S.predict(X)
            >>> plt.plot(X, y, label=r"$f(x)$", linestyle="dotted")
            >>> plt.scatter(X_train, y_train, label="Observations")
            >>> plt.plot(X, mean_prediction, label="Mean prediction")
            >>> plt.fill_between(
                X.ravel(),
                mean_prediction - 1.96 * std_prediction,
                mean_prediction + 1.96 * std_prediction,
                alpha=0.5,
                label=r"95% confidence interval",
                )
            >>> plt.legend()
            >>> plt.xlabel("$x$")
            >>> plt.ylabel("$f(x)$")
            >>> _ = plt.title("Gaussian process regression on noise-free dataset")

        """
        super().__init__(name, seed, log_level)

        self.noise = noise
        self.var_type = var_type
        self.cod_type = cod_type
        self.use_cod_y = use_cod_y
        self.name = name
        self.seed = seed
        self.log_level = log_level

        self.sigma = 0
        self.eps = sqrt(spacing(1))
        self.min_theta = min_theta
        self.max_theta = max_theta
        self.min_p = 1
        self.max_p = 2
        self.min_Lambda = 1e-9
        self.max_Lambda = 1.
        self.n_theta = n_theta
        self.n_p = n_p
        self.optim_p = optim_p
        # Psi matrix condition:
        self.cnd_Psi = 0
        self.inf_Psi = False

        self.model_optimizer = model_optimizer
        if self.model_optimizer is None:
            self.model_optimizer = differential_evolution
        self.model_fun_evals = model_fun_evals
        # differential evaluation uses maxiter = 1000
        # and sets the number of function evaluations to
        # (maxiter + 1) * popsize * N, which results in
        # 1000 * 15 * k, because the default popsize is 15 and
        # N is the number of parameters. This seems to be quite large:
        # for k=2 these are 30 000 iterations. Therefore we set this value to
        # 100
        if self.model_fun_evals is None:
            self.model_fun_evals = 100

        # Logging information
        self.log["negLnLike"] = []
        self.log["theta"] = []
        self.log["p"] = []
        self.log["Lambda"] = []
        # Logger
        logger.setLevel(self.log_level)
        logger.info(f"Starting the logger at level {self.log_level} for module {__name__}:")

    def exp_imp(self, y0, s0):
        """
        Returns the expected improvement for y0 and error s0 (in coded units).
        Args:
            y0 (float):
                function value (in coded units)
            s0 (float):
                error
        Returns:
            (float):
                The expected improvement value.
        """
        # y_min = min(self.cod_y)
        y_min = min(self.mean_cod_y)
        if s0 <= 0.0:
            EI = 0.0
        elif s0 > 0.0:
            EI_one = (y_min - y0) * (
                    0.5 + 0.5 * erf((1.0 / sqrt(2.0)) * ((y_min - y0) / s0))
            )
            EI_two = (s0 * (1.0 / sqrt(2.0 * pi))) * (
                exp(-(1.0 / 2.0) * ((y_min - y0) ** 2.0 / s0 ** 2.0))
            )
            EI = EI_one + EI_two
        return EI

    def set_de_bounds(self):
        """
        Determine search bounds for model_optimizer, e.g., differential evolution.
        """
        de_bounds = []
        for i in range(self.n_theta):
            de_bounds.append([self.min_theta, self.max_theta])
        if self.optim_p:
            for i in range(self.n_p):
                de_bounds.append([self.min_p, self.max_p])
            if self.noise:
                de_bounds.append([self.min_Lambda, self.max_Lambda])
        else:
            if self.noise:
                de_bounds.append([self.min_Lambda, self.max_Lambda])
        self.de_bounds = de_bounds

    def extract_from_bounds(self, new_theta_p_Lambda):
        """
        Extract `theta`, `p`, and `Lambda` from bounds. The kriging object stores
            `theta` as an array,  `p` as an array, and `Lambda` as a float.

        Args:
            new_theta_p_Lambda (numpy.array):
                1d-array with theta, p, and Lambda values. Order is important.

        """
        for i in range(self.n_theta):
            self.theta[i] = new_theta_p_Lambda[i]
        if self.optim_p:
            for i in range(self.n_p):
                self.p[i] = new_theta_p_Lambda[i + self.n_theta]
            if self.noise:
                self.Lambda = new_theta_p_Lambda[self.n_theta + self.n_p]
        else:
            if self.noise:
                self.Lambda = new_theta_p_Lambda[self.n_theta]

    def fit(self, nat_X, nat_y):
        """
        The function fits the hyperparameters (`theta`, `p`, `Lambda`) of the Kriging model, i.e.,
        the following internal values are computed:

        1. `theta`, `p`, and `Lambda` values via optimization of the function `fun_likelihood()`.
        2. Correlation matrix `Psi` via `rebuildPsi()`.

        Args:
            nat_X (array):
                sample points
            nat_y (array):
                function values

        Returns:
            surrogate (object):
                Fitted estimator.

        Attributes:
            theta (numpy.ndarray):
                Kriging theta values. Shape (k,).
            p (numpy.ndarray):
                Kriging p values. Shape (k,).
            LnDetPsi (numpy.float64):
                Determinant Psi matrix.
            Psi (numpy.matrix):
                Correlation matrix Psi. Shape (n,n).
            psi (numpy.ndarray):
                psi vector. Shape (n,).
            one (numpy.ndarray):
                vector of ones. Shape (n,).
            mu (numpy.float64):
                Kriging expected mean value mu.
            U (numpy.matrix):
                Kriging U matrix, Cholesky decomposition. Shape (n,n).
            SigmaSqr (numpy.float64):
                Sigma squared value.
            Lambda (float):
                lambda noise value.

        """
        self.nat_X = copy.deepcopy(nat_X)
        self.nat_y = copy.deepcopy(nat_y)
        self.n = self.nat_X.shape[0]
        self.k = self.nat_X.shape[1]
        self.cod_X = empty_like(self.nat_X)
        self.cod_y = empty_like(self.nat_y)

        # assume all variable types are "num" if "num" is
        # specified once:
        if len(self.var_type) < self.k:
            self.var_type = self.var_type * self.k
            logger.warning("Warning: All variable types forced to 'num'.")
        self.num_mask = array(list(map(lambda x: x == "num", self.var_type)))
        self.factor_mask = array(list(map(lambda x: x == "factor", self.var_type)))
        self.int_mask = array(list(map(lambda x: x == "int", self.var_type)))
        self.ordered_mask = array(list(map(lambda x: x == "int" or x == "num", self.var_type)))
        self.nat_to_cod_init()
        if self.n_theta > self.k:
            self.n_theta = self.k
            logger.warning("More theta values than dimensions. `n_theta` set to `k`.")
        self.theta = zeros(self.n_theta)
        # TODO: Currently not used:
        self.x0_theta = ones((self.n_theta,)) * self.n / (100 * self.k)
        self.p = ones(self.n_p) * 2.0

        self.pen_val = self.n * log(var(self.nat_y)) + 1e4
        self.negLnLike = None

        self.gen = spacefilling(k=self.k, seed=self.seed)

        # matrix related
        self.LnDetPsi = None
        self.Psi = zeros((self.n, self.n), dtype=float64)
        self.psi = zeros((self.n, 1))
        self.one = ones(self.n)
        self.mu = None
        self.U = None
        self.SigmaSqr = None
        self.Lambda = None
        # build_Psi() and build_U() are called in fun_likelihood
        self.set_de_bounds()
        if self.model_optimizer.__name__ == 'dual_annealing':
            result = self.model_optimizer(func=self.fun_likelihood,
                                          bounds=self.de_bounds)
        elif self.model_optimizer.__name__ == 'differential_evolution':
            result = self.model_optimizer(func=self.fun_likelihood,
                                          bounds=self.de_bounds,
                                          maxiter=self.model_fun_evals,
                                          seed=self.seed)
        elif self.model_optimizer.__name__ == 'direct':
            result = self.model_optimizer(func=self.fun_likelihood,
                                          bounds=self.de_bounds,
                                          # maxfun=self.model_fun_evals,
                                          eps=1e-2)
        elif self.model_optimizer.__name__ == 'shgo':
            result = self.model_optimizer(func=self.fun_likelihood,
                                          bounds=self.de_bounds)
        elif self.model_optimizer.__name__ == 'basinhopping':
            result = self.model_optimizer(func=self.fun_likelihood,
                                          x0=mean(self.de_bounds, axis=1))
        else:
            result = self.model_optimizer(func=self.fun_likelihood, bounds=self.de_bounds)
        # Finally, set new theta and p values and update the surrogate again
        # for new_theta_p_Lambda in de_results["x"]:
        new_theta_p_Lambda = result["x"]
        self.extract_from_bounds(new_theta_p_Lambda)
        self.build_Psi()
        self.build_U()
        # TODO: check if the following line is necessary!
        self.likelihood()
        self.log["negLnLike"] = append(self.log["negLnLike"], self.negLnLike)
        self.log["theta"] = append(self.log["theta"], self.theta)
        self.log["p"] = append(self.log["p"], self.p)
        self.log["Lambda"] = append(self.log["Lambda"], self.Lambda)
        # TODO: return self

    def fun_likelihood(self, new_theta_p_Lambda):
        """
        Compute log likelihood for a set of hyperparameters (theta, p, Lambda).
        Performs the following steps:

        1. Build Psi via `build_Psi()` and `build_U()`.
        2. Compute negLnLikelihood via `likelihood()
        3. If successful, the return `negLnLike` value, otherwise a penalty value (`pen_val`).

        Args:
            new_theta_p_Lambda (array):
                `theta`, `p`, and `Lambda` values stored in an array.
        Returns:
            (float):
                negLnLike, th negative log likelihood of the surface at the hyperparameters specified.
        """
        self.extract_from_bounds(new_theta_p_Lambda)
        if self.__is_any__(power(10.0, self.theta), 0):
            # print(f"Failure in fun_likelihood: 10^theta == 0. Setting negLnLike to {self.pen_val:.2f}.")
            logger.warning("Failure in fun_likelihood: 10^theta == 0. Setting negLnLike to %s", self.pen_val)
            return self.pen_val
        self.build_Psi()
        if (self.inf_Psi or self.cnd_Psi > 1e9):
            # print(f"\nFailure in fun_likelihood: Psi is ill conditioned ({self.cnd_Psi}).")
            # print(f"Setting negLnLike to {self.pen_val:.2f}.")
            logger.warning("Failure in fun_likelihood: Psi is ill conditioned: %s", self.cnd_Psi)
            logger.warning("Setting negLnLike to: %s", self.pen_val)
            return self.pen_val
        else:
            try:
                self.build_U()
            except Exception as err:
                f = self.pen_val
                print(f"Error in fun_likelihood(). Call to build_U() failed. {err=}, {type(err)=}")
                print(f"Setting negLnLike to {self.pen_val:.2f}.")
                return f
            self.likelihood()
            return self.negLnLike

    def __is_any__(self, x, v):
        if not isinstance(x, ndarray):
            x = array([x])
        return any(x == v)

    def build_Psi(self):
        """
        New construction (rebuild to reflect new data or a change in hyperparameters)
        of the (nxn) correlation matrix Psi as described in [Forr08a, p.57].
        Note:
            Method uses `theta`, `p`, and coded `X` values.
        """
        self.Psi = zeros((self.n, self.n), dtype=float64)
        theta = power(10.0, self.theta)
        if self.n_theta == 1:
            theta = theta * ones(self.k)
        try:
            D = zeros((self.n, self.n))
            if self.ordered_mask.any():
                X_ordered = self.cod_X[:, self.ordered_mask]
                D = squareform(
                    pdist(
                        X_ordered, metric='sqeuclidean', out=None, w=theta[self.ordered_mask]))
            if self.factor_mask.any():
                X_factor = self.cod_X[:, self.factor_mask]
                D = (D + squareform(
                    pdist(X_factor,
                          metric='hamming',
                          out=None,
                          w=theta[self.factor_mask])))
            self.Psi = exp(-D)
        except LinAlgError as err:
            print(f"Building Psi failed:\n {self.Psi}. {err=}, {type(err)=}")
        if self.noise:
            self.Psi = self.Psi + multiply(eye(self.n), self.Lambda)
        else:
            self.Psi = self.Psi + multiply(eye(self.n), self.eps)
        if (isinf(self.Psi)).any():
            self.inf_Psi = True
        self.cnd_Psi = cond(self.Psi)

    def build_U(self, scipy=True):
        """
        Cholesky factorization of Psi as U as described in [Forr08a, p.57].

        Args:
            scipy (bool): Use `scipy_cholesky`. If `False`, numpy's `cholesky` is used.
        """
        try:
            if scipy:
                self.U = scipy_cholesky(self.Psi, lower=True)
            else:
                self.U = cholesky(self.Psi)
            self.U = self.U.T
        except LinAlgError as err:
            print(f"build_U() Cholesky failed for Psi:\n {self.Psi}. {err=}, {type(err)=}")

    def likelihood(self):
        """
        Calculates the negative of the concentrated log-likelihood.
        Implementation of (2.32)  in [Forr08a].
        See also function krigingLikelihood() in spot.
        Note:
            `build_Psi` and `build_U` should be called first.
        Modifies:
            `mu`,
            `SigmaSqr`,
            `LnDetPsi`, and
            `negLnLike`, concentrated log-likelihood *-1 for minimizing

        """
        # (2.20) in [Forr08a]:
        mu = (
                 self.one.T.dot(
                     solve(self.U, solve(self.U.T, self.cod_y))
                 )
             ) / self.one.T.dot(solve(self.U, solve(self.U.T, self.one)))
        self.mu = mu
        # (2.31) in [Forr08a]
        self.SigmaSqr = (
                            (self.cod_y - self.one.dot(self.mu)).T.dot(
                                solve(
                                    self.U,
                                    solve(self.U.T, (self.cod_y - self.one.dot(self.mu))),
                                )
                            )
                        ) / self.n
        # (2.32) in [Forr08a]
        self.LnDetPsi = 2.0 * sum(log(abs(diag(self.U))))
        self.negLnLike = -1.0 * (-(self.n / 2.0) * log(self.SigmaSqr) - 0.5 * self.LnDetPsi)

    def plot(self, show=True):
        """
        This function plots 1d and 2d surrogates.
        Args:
            show (boolean):
                If `True`, the plots are displayed.
                If `False`, `plt.show()` should be called outside this function.
        """
        if self.k == 1:
            # TODO: Improve plot (add conf. interval etc.)
            fig = pylab.figure(figsize=(9, 6))
            # t1 = array(arange(0.0, 1.0, 0.01))
            # y1 = array([self.predict(array([x]), return_val="y") for x in t1])
            # plt.figure()
            # plt.plot(t1, y1, "k")
            # if show:
            #     plt.show()
            #
            n_grid = 100
            x = linspace(
                self.nat_range_X[0][0], self.nat_range_X[0][1], num=n_grid
            )
            y = self.predict(x)
            plt.figure()
            plt.plot(x, y, "k")
            if show:
                plt.show()

        if self.k == 2:
            fig = pylab.figure(figsize=(9, 6))
            n_grid = 100
            x = linspace(
                self.nat_range_X[0][0], self.nat_range_X[0][1], num=n_grid
            )
            y = linspace(
                self.nat_range_X[1][0], self.nat_range_X[1][1], num=n_grid
            )
            X, Y = meshgrid(x, y)
            # Predict based on the optimized results
            zz = array(
                [self.predict(array([x, y]), return_val="all") for x, y in zip(ravel(X), ravel(Y))]
            )
            zs = zz[:, 0, :]
            zse = zz[:, 1, :]
            Z = zs.reshape(X.shape)
            Ze = zse.reshape(X.shape)

            if self.cod_type == "norm":
                nat_point_X = (
                                      self.cod_X[:, 0] * (self.nat_range_X[0][1] - self.nat_range_X[0][0])
                              ) + self.nat_range_X[0][0]
                nat_point_Y = (
                                      self.cod_X[:, 1] * (self.nat_range_X[1][1] - self.nat_range_X[1][0])
                              ) + self.nat_range_X[1][0]
            elif self.cod_type == "std":
                nat_point_X = self.cod_X[:, 0] * self.nat_std_X[0] + self.nat_mean_X[0]
                nat_point_Y = self.cod_X[:, 1] * self.nat_std_X[1] + self.nat_mean_X[1]
            else:
                nat_point_X = self.cod_X[:, 0]
                nat_point_Y = self.cod_X[:, 1]
            contour_levels = 30
            ax = fig.add_subplot(224)
            # plot predicted values:
            pylab.contourf(X, Y, Ze, contour_levels, cmap="jet")
            pylab.title("Error")
            pylab.colorbar()
            # plot observed points:
            pylab.plot(nat_point_X, nat_point_Y, "ow")
            #
            ax = fig.add_subplot(223)
            # plot predicted values:
            plt.contourf(X, Y, Z, contour_levels, zorder=1, cmap="jet")
            plt.title("Surrogate")
            # plot observed points:
            pylab.plot(nat_point_X, nat_point_Y, "ow", zorder=3)
            pylab.colorbar()
            #
            ax = fig.add_subplot(221, projection="3d")
            ax.plot_surface(X, Y, Z, rstride=3, cstride=3, alpha=0.9, cmap="jet")
            #
            ax = fig.add_subplot(222, projection="3d")
            ax.plot_surface(X, Y, Ze, rstride=3, cstride=3, alpha=0.9, cmap="jet")
            #
            pylab.show()

    def predict(self, nat_X, nat=True, return_val="y"):
        """
        This function returns the prediction (in natural units) of the surrogate
        at the natural coordinates of X.

        Args:
            nat_X (array):
                Design variable to evaluate in natural units.
            nat (bool):
                argument `nat_X` is in natural range. Default: `True`. If set to `False`,
                `nat_X` will not be normalized (which might be useful if already normalized
                y values are used).
            return_val (string): whether `y`, `s`, neg. `ei` (negative expected improvement),
                or all three values are returned. Default is (for compatibility with sklearn) "y".
                To return `s`, select "s", to return neg. `ei`, select "ei".
                To return the tuple `(y, s, ei)`, select "all".
        Returns:
            (float):
                The predicted value in natural units.
            (float):
                predicted error
            (float):
                expected improvement
        """
        # Check for the shape and the type of the Input
        if isinstance(nat_X, ndarray):
            try:
                X = nat_X.reshape(-1, self.nat_X.shape[1])
                X = repair_non_numeric(X, self.var_type)
            except Exception:
                raise TypeError("13.1: Input to predict was not convertible to the size of X")
        else:
            raise TypeError(f"type of the given input is an {type(nat_X)} instead of an ndarray")

        # Iterate through the Input
        y = array([], dtype=float)
        s = array([], dtype=float)
        ei = array([], dtype=float)

        for i in range(X.shape[0]):
            # logger.debug(f"13.2: predict() Step 2: x (reshaped nat_X):\n {x}")
            if nat:
                x = self.nat_to_cod_x(X[i, :])
            else:
                x = X[i, :]
            y0, s0, ei0 = self.predict_coded(x)
            y = append(y, y0)
            s = append(s, s0)
            ei = append(ei, ei0)
        if return_val == "y":
            return y
        elif return_val == "s":
            return s
        elif return_val == "ei":
            return -1.0 * ei
        else:
            return y, s, -1.0 * ei

    def build_psi_vec(self, cod_x):
        """
        Build the psi vector. Needed by `predict_cod`, `predict_err_coded`,
        `regression_predict_coded`. Modifies `self.psi`.

        Args:
            cod_x (array): point to calculate psi

        """
        self.psi = zeros((self.n))
        # theta = self.theta  # TODO:
        theta = power(10.0, self.theta)
        if self.n_theta == 1:
            theta = theta * ones(self.k)
        try:
            D = zeros((self.n))
            if self.ordered_mask.any():
                X_ordered = self.cod_X[:, self.ordered_mask]
                x_ordered = cod_x[self.ordered_mask]
                D = cdist(x_ordered.reshape(-1, sum(self.ordered_mask)),
                          X_ordered.reshape(-1, sum(self.ordered_mask)),
                          metric='sqeuclidean',
                          out=None,
                          w=theta[self.ordered_mask])
            if self.factor_mask.any():
                X_factor = self.cod_X[:, self.factor_mask]
                x_factor = cod_x[self.factor_mask]
                D = (D + cdist(x_factor.reshape(-1, sum(self.factor_mask)),
                               X_factor.reshape(-1, sum(self.factor_mask)),
                               metric='hamming',
                               out=None,
                               w=theta[self.factor_mask]))
            self.psi = exp(-D).T
        except LinAlgError as err:
            print(f"Building psi failed:\n {self.psi}. {err=}, {type(err)=}")

    def predict_coded(self, cod_x):
        """
        Kriging prediction of one point in the coded units as described in (2.20) in [Forr08a].
        The error is returned as well.
        See also [Forr08a, p.60].
        Note:
            `self.mu` and `self.SigmaSqr` are computed in `likelihood`, not here.
        Args:
           cod_x (array):
            point in coded units to make prediction at
        Returns:
            (float):
                predicted value in coded units.
            (float):
                predicted error.
        """
        self.build_psi_vec(cod_x)
        f = self.mu + self.psi.T.dot(
            solve(self.U, solve(
                self.U.T, self.cod_y - self.one.dot(self.mu)))
        )
        try:
            if self.noise:
                Lambda = self.Lambda
            else:
                Lambda = 0.0
            # Error in [Forr08a, p.87]:
            SSqr = self.SigmaSqr * (1 + Lambda - self.psi.T.dot(solve(self.U, solve(self.U.T, self.psi))))
        except Exception as err:
            print(f"Could not determine SSqr. Wrong or missing Lambda? {err=}, {type(err)=}")

        SSqr = power(abs(SSqr[0]), 0.5)[0]
        EI = self.exp_imp(y0=f[0], s0=SSqr)
        return f[0], SSqr, EI

    def weighted_exp_imp(self, cod_x, w):
        """
        Weighted expected improvement.

        References:
            [Sobester et al. 2005].

        Args:
            cod_x (array):
                A coded design vector.
            w (float):
                weight

        Returns:
            (float):
                weighted expected improvement.
        """
        y0, s0 = self.predict_coded(cod_x)
        y_min = min(self.cod_y)
        if s0 <= 0.0:
            EI = 0.0
        elif s0 > 0.0:
            EI_one = w * (
                    (y_min - y0)
                    * (0.5 + 0.5 * erf((1.0 / sqrt(2.0)) * ((y_min - y0) / s0)))
            )
            EI_two = (
                    (1.0 - w)
                    * (s0 * (1.0 / sqrt(2.0 * pi)))
                    * (exp(-(1.0 / 2.0) * ((y_min - y0) ** 2.0 / s0 ** 2.0)))
            )
            EI = EI_one + EI_two
        return EI

    def calculate_mean_MSE(self, n_samples=200, points=None):
        """
        This function calculates the mean MSE metric of the model by evaluating MSE at a number of points.
        Args:
            n_samples (integer):
                Number of points to sample the mean squared error at.
                Ignored if the points argument is specified.
            points (array):
                an array of points to sample the model at.
        Returns:
            (float):
                the mean value of MSE and the standard deviation of the MSE points
        """
        if points is None:
            points = self.gen.lhd(n_samples)
        values = zeros(len(points))
        for enu, point in enumerate(points):
            s0 = self.predict(cod_X=point, nat=True, return_val="s")
            values[enu] = s0
        return mean(values), std(values)

    def cod_to_nat_x(self, cod_X):
        """
        Args:
            cod_X (array):
                An array representing one point (self.k long) in normalized (coded) units.
        Returns:
            (array):
                An array of natural (physical or real world) units.
        """
        X = copy.deepcopy(cod_X)
        if self.cod_type == "norm":
            for i in range(self.k):
                X[i] = (
                               X[i] * float(self.nat_range_X[i][1] - self.nat_range_X[i][0])
                       ) + self.nat_range_X[i][0]
            return X
        elif self.cod_type == "std":
            for i in range(self.k):
                X[i] = X[i] * self.nat_std_X[i] + self.nat_mean_X[i]
            return X
        else:
            return cod_X

    def cod_to_nat_y(self, cod_y):
        """
        Args:
            cod_y (array):
                A normalized array of coded (model) units in the range of [0,1].
        Returns:
            (array):
                An array of observed values in real-world units.
        """
        if self.cod_type == "norm":
            return (
                           cod_y * (self.nat_range_y[1] - self.nat_range_y[0])
                   ) + self.nat_range_y[0]
        elif self.cod_type == "std":
            return cod_y * self.nat_std_y + self.nat_mean_y
        else:
            return cod_y

    def nat_to_cod_x(self, nat_X):
        """
        Normalize one point (row) of nat_X array to [0,1]. The internal nat_range_X values are not updated.

        Args:
            nat_X (array):
                An array representing one points (self.k long) in natural (physical or real world) units.
        Returns:
            (array):
                An array of coded values in the range of [0,1] for each dimension.
        """
        X = copy.deepcopy(nat_X)
        if self.cod_type == "norm":
            for i in range(self.k):
                # TODO: Check Implementation of range correction if range == 0:
                # rangex <- xmax - xmin
                # rangey <- ymax - ymin
                # xmin[rangex == 0] <- xmin[rangex == 0] - 0.5
                # xmax[rangex == 0] <- xmax[rangex == 0] + 0.5
                # rangex[rangex == 0] <- 1
                # logger.debug(f"self.nat_range_X[{i}]:\n {self.nat_range_X[i]}")
                # logger.debug(f"X[{i}]:\n {X[i]}")
                rangex = float(self.nat_range_X[i][1] - self.nat_range_X[i][0])
                if rangex == 0:
                    self.nat_range_X[i][0] = self.nat_range_X[i][0] - 0.5
                    self.nat_range_X[i][1] = self.nat_range_X[i][1] + 0.5
                X[i] = (X[i] - self.nat_range_X[i][0]) / float(self.nat_range_X[i][1] - self.nat_range_X[i][0])
            return X
        elif self.cod_type == "std":
            for i in range(self.k):
                X[i] = (X[i] - self.nat_mean_X[i]) / self.nat_std_X[i]
            return X
        else:
            return nat_X

    def nat_to_cod_y(self, nat_y):
        """
        Normalize natural y values to [0,1].
        Args:
            nat_y (array):
                An array of observed values in natural (real-world) units.
        Returns:
            (array):
                A normalized array of coded (model) units in the range of [0,1].

        """
        if self.use_cod_y:
            if self.cod_type == "norm":
                return (nat_y - self.nat_range_y[0]) / (self.nat_range_y[1] - self.nat_range_y[0])
            elif self.cod_type == "std":
                return (nat_y - self.nat_mean_y) / self.nat_std_y
            else:
                return nat_y
        else:
            return nat_y

    def nat_to_cod_init(self):
        """
        Determine max and min of each dimension and normalize that axis to a range of [0,1].
        Called when 1) surrogate is initialized and 2) new points arrive, i.e., suggested
        by the surrogate as infill points.
        This method calls `nat_to_cod_x` and `nat_to_cod_y` and updates the ranges `nat_range_X` and
        `nat_range_y`.
        """
        self.nat_range_X = []
        self.nat_range_y = []
        for i in range(self.k):
            self.nat_range_X.append([min(self.nat_X[:, i]), max(self.nat_X[:, i])])
        self.nat_range_y.append(min(self.nat_y))
        self.nat_range_y.append(max(self.nat_y))
        self.nat_mean_X = mean(self.nat_X, axis=0)
        self.nat_std_X = std(self.nat_X, axis=0)
        self.nat_mean_y = mean(self.nat_y)
        self.nat_std_y = std(self.nat_y)
        Z = aggregate_mean_var(X=self.nat_X, y=self.nat_y)
        mu = Z[1]
        self.mean_cod_y = empty_like(mu)

        for i in range(self.n):
            self.cod_X[i] = self.nat_to_cod_x(self.nat_X[i])
        for i in range(self.n):
            self.cod_y[i] = self.nat_to_cod_y(self.nat_y[i])
        for i in range(mu.shape[0]):
            self.mean_cod_y[i] = self.nat_to_cod_y(mu[i])



import numpy as np


def selectNew(A, X, tolerance=0):
    """
    Select rows from A that are not in X.

    Args:
        A (numpy.ndarray): A array with new values
        X (numpy.ndarray): X array with known values

    Returns:
        (numpy.ndarray): array with unknown (new) values
        (numpy.ndarray): array with `True` if value is new, otherwise `False`.
    """
    ind = np.zeros(A.shape[0], dtype=bool)
    for i in range(X.shape[0]):
        B = np.abs(A - X[i, :])
        ind = ind + np.all(B <= tolerance, axis=1)
    return A[~ind], ~ind



def test_repair_non_numeric():
    """
    Test repair_non_numeric method
    """
    import numpy as np
    from spotPython.fun.objectivefunctions import analytical
    from spotPython.spot.spot import Spot
    from spotPython.utils.repair import repair_non_numeric

    fun = analytical().fun_branin_factor
    ni=12
    spot_test = Spot(
        fun=fun,
        lower=np.array([-5,-0,1]),
        upper=np.array([10,15,3]),
        var_type=["num", "num", "factor"],
        design_control={"init_size": ni}
    )
    spot_test.run()
    # 3rd variable should be a rounded float, because it was labeled as a factor
    assert spot_test.min_X[2] == round(spot_test.min_X[2])

    spot_test.X = spot_test.generate_design(size=spot_test.design_control["init_size"],
                                            repeats=spot_test.design_control["repeats"],
                                            lower=spot_test.lower,
                                            upper=spot_test.upper)
    spot_test.X = repair_non_numeric(spot_test.X,
                                        spot_test.var_type)
    assert spot_test.X.ndim == 2
    assert spot_test.X.shape[0] == ni
    assert spot_test.X.shape[1] == 3



def one_hot_encode(strings) -> dict:
    """One hot encode a list of strings.
    Arguments:
        strings (list): List of strings to encode.
    Returns:
        dict: Dictionary of strings and their one hot encoded values.
    Example:
        >>> one_hot_encode(['a', 'b', 'c'])
        {'a': [1, 0, 0], 'b': [0, 1, 0], 'c': [0, 0, 1]}
    """
    n = len(strings)
    encoding_dict = {}
    for i, string in enumerate(strings):
        one_hot_encoded_value = [0] * n
        one_hot_encoded_value[i] = 1
        encoding_dict[string] = one_hot_encoded_value
    return encoding_dict


def sum_encoded_values(strings, encoding_dict) -> int:
    """Sum the encoded values of a list of strings.

    Args:
        strings (list): List of strings to encode.
        encoding_dict (dict): Dictionary of strings and their one hot encoded values.

    Returns:
        int: Decimal value of the sum of the encoded values.

    Example:
        >>> encoding_dict = {'a': [1, 0, 0], 'b': [0, 1, 0], 'c': [0, 0, 1]}
            sum_encoded_values(['a', 'b', 'c'], encoding_dict)
            7
            sum_encoded_values(['a', 'c'], encoding_dict)
            5
    """
    result = [0] * len(list(encoding_dict.values())[0])
    for string in strings:
        encoded_value = encoding_dict.get(string)
        if encoded_value:
            result = [sum(x) for x in zip(result, encoded_value)]
    decimal_result = 0
    for i, value in enumerate(result[::-1]):
        decimal_result += value * (2**i)
    return decimal_result


import json


def get_one_hot(alg: str, hyper_param: str, d: dict = None, filename: str = "data.json") -> dict:
    """Get one hot encoded values for a hyper parameter of an algorithm.
    Arguments:
        alg (str): Name of the algorithm.
        hyper_param (str): Name of the hyper parameter.
        d (dict): Dictionary of algorithms and their hyperparameters.
        filename (str): Name of the file containing the dictionary.
    Returns:
        dict: Dictionary of hyper parameter values and their one hot encoded values.
    Example:
        >>> alg = "HoeffdingAdaptiveTreeClassifier"
            hyper_param = "split_criterion"
            d = {
                "HoeffdingAdaptiveTreeClassifier": {
                    "split_criterion": ["gini", "info_gain", "hellinger"],
                    "leaf_prediction": ["mc", "nb", "nba"],
                    "bootstrap_sampling": ["0", "1"]
                    },
                    "HoeffdingTreeClassifier": {
                        "split_criterion": ["gini", "info_gain", "hellinger"],
                        "leaf_prediction": ["mc", "nb", "nba"],
                        "binary_split": ["0", "1"],
                        "stop_mem_management": ["0", "1"]
                    }
                }
            get_one_hot(alg, hyper_param, d)
            {'gini': [1, 0, 0], 'info_gain': [0, 1, 0], 'hellinger': [0, 0, 1]}
    """
    if d is None:
        with open(filename, "r") as f:
            d = json.load(f)
    values = d[alg][hyper_param]
    one_hot_encoded_values = one_hot_encode(values)
    return one_hot_encoded_values


def add_missing_elements(a: list, b: list) -> list:
    """Add missing elements from list a to list b.
    Arguments:
        a (list): List of elements to check.
        b (list): List of elements to add to.

    Returns:
        list: List of elements with missing elements from list a added.

    Example:
        >>> a = [1, 4]
            b = [1, 2]
            add_missing_elements(a, b)
            [1, 2, 4]
    """
    for element in a:
        if element not in b:
            b.append(element)


def find_closest_key(integer_value: int, encoding_dict: dict) -> str:
    """
    Given an integer value and an encoding dictionary that maps keys to binary values,
    this function finds the key in the dictionary whose binary value is closest to the binary representation of the integer value.

    Arguments:
        integer_value (int): The integer value to find the closest key for.
        encoding_dict (dict): The encoding dictionary that maps keys to binary values.
    Returns:
        str: The key in the encoding dictionary whose binary value is closest to the binary representation of the integer value.
    Example:
        >>> encoding_dict = {'A': [1, 0, 0], 'B': [0, 1, 0], 'C': [0, 0, 1]}
            find_closest_key(6, encoding_dict)
            'B'
    """
    binary_value = [int(x) for x in format(integer_value, f"0{len(list(encoding_dict.values())[0])}b")]
    min_distance = float("inf")
    closest_key = None
    for key, encoded_value in encoding_dict.items():
        distance = sum([x != y for x, y in zip(binary_value, encoded_value)])
        if distance < min_distance:
            min_distance = distance
            closest_key = key
    return closest_key



import numpy as np
from numpy.random import default_rng
from random import random


class analytical:
    """
    Analytical test functions.

    Args:
        offset (float): offset
        hz (float): hz
        seed (int): seed.
            See [Numpy Random Sampling](https://numpy.org/doc/stable/reference/random/index.html#random-quick-start)

    """

    def __init__(self, offset=0.0, hz=0, seed=126):
        self.offset = offset
        self.hz = hz
        self.seed = seed
        self.rng = default_rng(seed=self.seed)
        self.fun_control = {"sigma": 0, "seed": None, "sel_var": None}

    def add_noise(self, y):
        # Use own rng:
        if self.fun_control["seed"] is not None:
            rng = default_rng(seed=self.fun_control["seed"])
        # Use class rng:
        else:
            rng = self.rng
        noise_y = np.array([], dtype=float)
        for y_i in y:
            noise_y = np.append(
                noise_y,
                y_i + rng.normal(loc=0, scale=self.fun_control["sigma"], size=1),
            )
        return noise_y

    def fun_linear(self, X, fun_control=None):
        """Linear function.

        Args:
            X (array): input

        Returns:
            (float): objective function value.
        """
        if fun_control is not None:
            self.fun_control = fun_control
        try:
            X.shape[1]
        except ValueError as err:
            print("error message:", err)
            X = np.array(X)

        if len(X.shape) < 2:
            X = np.array([X])
        y = np.array([], dtype=float)
        for i in range(X.shape[0]):
            y = np.append(y, np.sum(X[i]))
        if self.fun_control["sigma"] > 0:
            return self.add_noise(y)
        else:
            return y

    def fun_sphere(self, X, fun_control=None):
        """Sphere function.

        Args:
            X (array): input
            fun_control (dict): dict with entries `seed` and `sigma`.

        Returns:
            (float): function values
        """
        if fun_control is not None:
            self.fun_control = fun_control
        try:
            X.shape[1]
        except ValueError:
            X = np.array(X)

        if len(X.shape) < 2:
            X = np.array([X])
        offset = np.ones(X.shape[1]) * self.offset
        y = np.array([], dtype=float)
        for i in range(X.shape[0]):
            y = np.append(y, np.sum((X[i] - offset) ** 2))
        # TODO: move to a separate function:
        if self.fun_control["sigma"] > 0:
            # Use own rng:
            if self.fun_control["seed"] is not None:
                rng = default_rng(seed=fun_control["seed"])
            # Use class rng:
            else:
                rng = self.rng
            noise_y = np.array([], dtype=float)
            for y_i in y:
                noise_y = np.append(noise_y, y_i + rng.normal(loc=0, scale=fun_control["sigma"], size=1))
            return noise_y
        else:
            return y

    def fun_cubed(self, X, fun_control=None):
        if fun_control is None:
            fun_control = self.fun_control
        try:
            X.shape[1]
        except ValueError:
            X = np.array(X)

        if len(X.shape) < 2:
            X = np.array([X])
        offset = np.ones(X.shape[1]) * self.offset
        y = np.array([], dtype=float)
        for i in range(X.shape[0]):
            y = np.append(y, np.sum((X[i] - offset) ** 3))
        # TODO: move to a separate function:
        if fun_control["sigma"] > 0:
            # Use own rng:
            if fun_control["seed"] is not None:
                rng = default_rng(seed=fun_control["seed"])
            # Use class rng:
            else:
                rng = self.rng
            noise_y = np.array([], dtype=float)
            for i in y:
                # noise_y = np.append(
                #     noise_y, i + np.random.normal(loc=0, scale=self.sigma, size=1)
                noise_y = np.append(noise_y, i + rng.normal(loc=0, scale=fun_control["sigma"], size=1))
            return noise_y
        else:
            return y

    def fun_forrester(self, X, fun_control=None):
        """
        Function used by [Forr08a, p.83].
        f(x) = (6x- 2)^2 sin(12x-4) for x in [0,1].
        Starts with three sample points at x=0, x=0.5, and x=1.

        Args:
            X (flooat): input values (1-dim)

        Returns:
            (float): function value
        """
        if fun_control is None:
            fun_control = self.fun_control
        try:
            X.shape[1]
        except ValueError:
            X = np.array(X)

        if len(X.shape) < 2:
            X = np.array([X])
        y = np.array([], dtype=float)
        for i in range(X.shape[0]):
            y = np.append(y, (6.0 * X[i] - 2) ** 2 * np.sin(12 * X[i] - 4))
        # TODO: move to a separate function:
        if fun_control["sigma"] > 0:
            # Use own rng:
            if fun_control["seed"] is not None:
                rng = default_rng(seed=fun_control["seed"])
            # Use class rng:
            else:
                rng = self.rng
            noise_y = np.array([], dtype=float)
            for i in y:
                # noise_y = np.append(
                #     noise_y, i + np.random.normal(loc=0, scale=self.sigma, size=1)
                noise_y = np.append(noise_y, i + rng.normal(loc=0, scale=fun_control["sigma"], size=1))
            return noise_y
        else:
            return y

    def fun_branin(self, X, fun_control=None):
        """Branin function.

        The 2-dim Branin function is
        y = a * (x2 - b * x1**2 + c * x1 - r) ** 2 + s * (1 - t) * np.cos(x1) + s,
        where values of a, b, c, r, s and t are: a = 1, b = 5.1 / (4*pi**2),
        c = 5 / pi, r = 6, s = 10 and t = 1 / (8*pi).

        It has three global minima:
        f(x) = 0.397887 at (-pi, 12.275), (pi, 2.275), and (9.42478, 2.475).

        Input Domain:
        This function is usually evaluated on the square  x1 in  [-5, 10] x x2 in [0, 15].

        Args:
            X (array): input value
            fun_control (dict): dict with entries `seed` and `sigma`.

        Returns:
            (float): function value

        """
        if fun_control is None:
            fun_control = self.fun_control
        try:
            X.shape[1]
        except ValueError:
            X = np.array([X])
        if X.shape[1] != 2:
            raise Exception
        x1 = X[:, 0]
        x2 = X[:, 1]
        a = 1
        b = 5.1 / (4 * np.pi**2)
        c = 5 / np.pi
        r = 6
        s = 10
        t = 1 / (8 * np.pi)
        y = a * (x2 - b * x1**2 + c * x1 - r) ** 2 + s * (1 - t) * np.cos(x1) + s
        # TODO: move to a separate function:
        if fun_control["sigma"] > 0:
            # Use own rng:
            if fun_control["seed"] is not None:
                rng = default_rng(seed=fun_control["seed"])
            # Use class rng:
            else:
                rng = self.rng
            noise_y = np.array([], dtype=float)
            for i in y:
                # noise_y = np.append(
                #     noise_y, i + np.random.normal(loc=0, scale=self.sigma, size=1)
                noise_y = np.append(noise_y, i + rng.normal(loc=0, scale=fun_control["sigma"], size=1))
            return noise_y
        else:
            return y

    def fun_branin_factor(self, X, fun_control=None):
        """Branin function with factor variable x_3.

        The 2-dim Branin, or Branin-Hoo, function has three global minima.
        The recommended values of a, b, c, r, s and t are: a = 1, b = 5.1 / (4*pi**2),
        c = 5 / π, r = 6, s = 10 and t = 1 / (8*pi).

        Input Domain:
        This function is usually evaluated on the square  x1 in  [-5, 10] x x2 in [0, 15]
        and with x3 from the set {0, 1, 2}, i.e., x3 is a factor variable with three levels.

        Global Minimum:
        f(x) = 0.397887 -1  at (-pi, 12.275, 2), (pi, 2.275, 2), and (9.42478, 2.475, 2).

        Args:
            X (array): input value

        Returns:
            (float): function value

        """
        if fun_control is None:
            fun_control = self.fun_control
        try:
            X.shape[1]
        except ValueError:
            X = np.array([X])
        if X.shape[1] != 3:
            raise Exception
        x1 = X[:, 0]
        x2 = X[:, 1]
        x3 = X[:, 2]
        a = 1
        b = 5.1 / (4 * np.pi**2)
        c = 5 / np.pi
        r = 6
        s = 10
        t = 1 / (8 * np.pi)
        y = a * (x2 - b * x1**2 + c * x1 - r) ** 2 + s * (1 - t) * np.cos(x1) + s
        for j in range(X.shape[0]):
            if x3[j] == 1:
                y[j] = y[j] + 10
            elif x3[j] == 2:
                y[j] = y[j] - 10
        # TODO: move to a separate function:
        if fun_control["sigma"] > 0:
            # Use own rng:
            if fun_control["seed"] is not None:
                rng = default_rng(seed=fun_control["seed"])
            # Use class rng:
            else:
                rng = self.rng
            noise_y = np.array([], dtype=float)
            for i in y:
                noise_y = np.append(noise_y, i + rng.normal(loc=0, scale=fun_control["sigma"], size=1))
            return noise_y
        else:
            return y

    def fun_branin_modified(self, X, fun_control=None):
        if fun_control is None:
            fun_control = self.fun_control
        try:
            X.shape[1]
        except ValueError:
            X = np.array([X])

        if X.shape[1] != 2:
            raise Exception
        x = X[:, 0]
        y = X[:, 1]
        X1 = 15 * x - 5
        X2 = 15 * y
        a = 1
        b = 5.1 / (4 * np.pi**2)
        c = 5 / np.pi
        d = 6
        e = 10
        ff = 1 / (8 * np.pi)
        y = (a * (X2 - b * X1**2 + c * X1 - d) ** 2 + e * (1 - ff) * np.cos(X1) + e) + 5 * x
        # TODO: move to a separate function:
        if fun_control["sigma"] > 0:
            # Use own rng:
            if fun_control["seed"] is not None:
                rng = default_rng(seed=fun_control["seed"])
            # Use class rng:
            else:
                rng = self.rng
            noise_y = np.array([], dtype=float)
            for i in y:
                noise_y = np.append(noise_y, i + rng.normal(loc=0, scale=fun_control["sigma"], size=1))
            return noise_y
        else:
            return y

    def branin_noise(self, X):
        try:
            X.shape[1]
        except ValueError:
            X = np.array([X])

        if X.shape[1] != 2:
            raise Exception
        x = X[:, 0]
        y = X[:, 1]
        X1 = 15 * x - 5
        X2 = 15 * y
        a = 1
        b = 5.1 / (4 * np.pi**2)
        c = 5 / np.pi
        d = 6
        e = 10
        ff = 1 / (8 * np.pi)
        noiseFree = (a * (X2 - b * X1**2 + c * X1 - d) ** 2 + e * (1 - ff) * np.cos(X1) + e) + 5 * x
        noise_y = []
        for i in noiseFree:
            noise_y.append(i + np.random.standard_normal() * 15)
        return np.array(noise_y)

    def fun_sin_cos(self, X, fun_control=None):
        if fun_control is None:
            fun_control = self.fun_control
        try:
            X.shape[1]
        except ValueError:
            X = np.array([X])
        if X.shape[1] != 2:
            raise Exception
        x0 = X[:, 0]
        x1 = X[:, 1]
        y = 2.0 * np.sin(x0 + self.hz) + 0.5 * np.cos(x1 + self.hz)
        # TODO: move to a separate function:
        if fun_control["sigma"] > 0:
            # Use own rng:
            if fun_control["seed"] is not None:
                rng = default_rng(seed=fun_control["seed"])
            # Use class rng:
            else:
                rng = self.rng
            noise_y = np.array([], dtype=float)
            for i in y:
                noise_y = np.append(noise_y, i + rng.normal(loc=0, scale=fun_control["sigma"], size=1))
            return noise_y
        else:
            return y

    # def fun_forrester_2(self, X):
    #     """
    #     Function used by [Forr08a, p.83].
    #     f(x) = (6x- 2)^2 sin(12x-4) for x in [0,1].
    #     Starts with three sample points at x=0, x=0.5, and x=1.

    #     Args:
    #         X (flooat): input values (1-dim)

    #     Returns:
    #         float: function value
    #     """
    #     try:
    #         X.shape[1]
    #     except ValueError:
    #         X = np.array(X)

    #     if len(X.shape) < 2:
    #         X = np.array([X])
    #     # y = X[:, 1]
    #     y = (6.0 * X - 2) ** 2 * np.sin(12 * X - 4)
    #     if self.sigma != 0:
    #         noise_y = np.array([], dtype=float)
    #         for i in y:
    #             noise_y = np.append(
    #                 noise_y, i + np.random.normal(loc=0, scale=self.sigma, size=1)
    #             )
    #         return noise_y
    #     else:
    #         return y

    def fun_runge(self, X, fun_control=None):
        """
        Runge function.
        Formula: f(x) = 1/ (1 + sum(x_i) - offset)^2
        Dim: k >= 1
        Interval: -5 <= x <= 5

        Args:
            X (numpy.array):
            input
            fun_control (dictionary, optional):
            control parameters. Defaults to None.

        Returns:
            (float) :
            function value
        """
        if fun_control is None:
            fun_control = self.fun_control
        try:
            X.shape[1]
        except ValueError:
            X = np.array(X)

        if len(X.shape) < 2:
            X = np.array([X])
        offset = np.ones(X.shape[1]) * self.offset
        y = np.array([], dtype=float)
        for i in range(X.shape[0]):
            y = np.append(y, (1 / (1 + np.sum((X[i] - offset) ** 2))))
        # TODO: move to a separate function:
        if fun_control["sigma"] > 0:
            # Use own rng:
            if fun_control["seed"] is not None:
                rng = default_rng(seed=fun_control["seed"])
            # Use class rng:
            else:
                rng = self.rng
            noise_y = np.array([], dtype=float)
            for i in y:
                noise_y = np.append(noise_y, i + rng.normal(loc=0, scale=fun_control["sigma"], size=1))
            return noise_y
        else:
            return y

    def fun_wingwt(self, X, fun_control=None):
        """
        Wing weight function. Example from Forrester et al. to understand the weight
        of an unpainted light aircraft wing as a function of nine design and operational parameters:
        W = 0.036 S_W**0.758 * Wfw**0.0035 ( A / (cos**2 Lambda))**0.6 *
            q**0.006  * lambda**0.04 * ( (100 Rtc)/(cos Lambda) ))**-0.3*
            (Nz Wdg)**0.49

        | Symbol    | Parameter                              | Baseline | Minimum | Maximum |
        |-----------|----------------------------------------|----------|---------|---------|
        | $S_W$     | Wing area ($ft^2$)                     | 174      | 150     | 200     |
        | $W_{fw}$  | Weight of fuel in wing (lb)            | 252      | 220     | 300     |
        | $A$       | Aspect ratio                          | 7.52     | 6       | 10      |
        | $Lambda$ | Quarter-chord sweep (deg)              | 0        | -10     | 10      |
        | $q$       | Dynamic pressure at cruise ($lb/ft^2$) | 34       | 16      | 45      |
        | $lambda$ | Taper ratio                            | 0.672    | 0.5     | 1       |
        | $R_{tc}$  | Aerofoil thickness to chord ratio      | 0.12     | 0.08    | 0.18    |
        | $N_z$     | Ultimate load factor                   | 3.8      | 2.5     | 6       |
        | $W_{dg}$  | Flight design gross weight (lb)         | 2000     | 1700    | 2500    |
        | $W_p$     | paint weight (lb/ft^2)                   | 0.064 |   0.025  | 0.08    |

        Args:
            X (numpy.array):
                10-dim input vector
            fun_control (dictionary, optional):
                control parameters. Defaults to None.

        Returns:
            (float) :
            function value
        """
        if fun_control is None:
            fun_control = self.fun_control
        try:
            X.shape[1]
        except ValueError:
            X = np.array(X)
        #
        W_res = np.array([], dtype=float)
        for i in range(X.shape[0]):
            Sw = X[i, 0] * (200 - 150) + 150
            Wfw = X[i, 1] * (300 - 220) + 220
            A = X[i, 2] * (10 - 6) + 6
            L = (X[i, 3] * (10 - (-10)) - 10) * np.pi / 180
            q = X[i, 4] * (45 - 16) + 16
            la = X[i, 5] * (1 - 0.5) + 0.5
            Rtc = X[i, 6] * (0.18 - 0.08) + 0.08
            Nz = X[i, 7] * (6 - 2.5) + 2.5
            Wdg = X[i, 8] * (2500 - 1700) + 1700
            Wp = X[i, 9] * (0.08 - 0.025) + 0.025
            # calculation on natural scale
            W = 0.036 * Sw**0.758 * Wfw**0.0035 * (A / np.cos(L) ** 2) ** 0.6 * q**0.006
            W = W * la**0.04 * (100 * Rtc / np.cos(L)) ** (-0.3) * (Nz * Wdg) ** (0.49) + Sw * Wp
            W_res = np.append(W_res, W)
        return W_res

    def fun_xsin(self, X, fun_control=None):
        """
        Args:
            X (float): input values (1-dim)

        Returns:
            (float): function value
        """
        if fun_control is None:
            fun_control = self.fun_control
        try:
            X.shape[1]
        except ValueError:
            X = np.array(X)

        if len(X.shape) < 2:
            X = np.array([X])
        y = np.array([], dtype=float)
        for i in range(X.shape[0]):
            y = np.append(y, X[i] * np.sin(1.0 / X[i]))
        # TODO: move to a separate function:
        if fun_control["sigma"] > 0:
            # Use own rng:
            if fun_control["seed"] is not None:
                rng = default_rng(seed=fun_control["seed"])
            # Use class rng:
            else:
                rng = self.rng
            noise_y = np.array([], dtype=float)
            for i in y:
                # noise_y = np.append(
                #     noise_y, i + np.random.normal(loc=0, scale=self.sigma, size=1)
                noise_y = np.append(noise_y, i + rng.normal(loc=0, scale=fun_control["sigma"], size=1))
            return noise_y
        else:
            return y

    def fun_rosen(self, X, fun_control=None):
        if fun_control is None:
            fun_control = self.fun_control
        try:
            X.shape[1]
        except ValueError:
            X = np.array([X])
        if X.shape[1] != 2:
            raise Exception
        x0 = X[:, 0]
        x1 = X[:, 1]
        b = 10
        y = (x0 - 1) ** 2 + b * (x1 - x0**2) ** 2
        if self.fun_control["sigma"] > 0:
            return self.add_noise(y)
        else:
            return y

    def fun_random_error(self, X, fun_control=None):
        """Return errors for testing spot stability.

        Args:
            X (array): input

        Returns:
            (float): objective function value.
        """
        if fun_control is not None:
            self.fun_control = fun_control
        try:
            X.shape[1]
        except ValueError as err:
            print("error message:", err)
            X = np.array(X)
        if len(X.shape) < 2:
            X = np.array([X])
        y = np.array([], dtype=float)
        for i in range(X.shape[0]):
            # provoke error:
            if random() < 0.1:
                y = np.append(y, np.nan)
            else:
                y = np.append(y, np.sum(X[i]))
        if self.fun_control["sigma"] > 0:
            return self.add_noise(y)
        else:
            print(y)
            return y



def test_show_progress():
    """
    Test show_progress
    """
    import numpy as np
    from spotPython.fun.objectivefunctions import analytical
    from spotPython.spot import spot
    import numpy as np
    from math import inf
    # number of initial points:
    ni = 7
    # number of points
    n = 10

    fun = analytical().fun_sphere
    lower = np.array([-1])
    upper = np.array([1])
    design_control={"init_size": ni}

    spot_1 = spot.Spot(fun=fun,
                lower = lower,
                upper= upper,
                fun_evals = n,
                show_progress=False,
                design_control=design_control,)
    spot_1.run()
    # To check whether the run was successfully completed,
    # we compare the number of evaluated points to the specicified
    # number of points.
    assert spot_1.y.shape[0] == n

    spot_2 = spot.Spot(fun=fun,
                lower = lower,
                upper= upper,
                fun_evals = n,
                show_progress=True,
                design_control=design_control,)
    spot_2.run()
    # To check whether the run was successfully completed,
    # we compare the number of evaluated points to the specicified
    # number of points.
    assert spot_2.y.shape[0] == n

    spot_3 = spot.Spot(fun=fun,
                lower=lower,
                upper=upper,
                fun_evals=inf,
                max_time=0.1,
                show_progress=False,
                design_control=design_control,)
    spot_3.run()
    # To check whether the run was successfully completed,
    # we test whether the number of evaluated points is greater than zero,
    # because we do not know how many points can be evaluated.
    assert spot_3.y.shape[0] > 0

    spot_4 = spot.Spot(fun=fun,
                lower=lower,
                upper=upper,
                fun_evals=inf,
                max_time=0.1,
                show_progress=True,
                design_control=design_control,)
    spot_4.run()
    # To check whether the run was successfully completed,
    # we test whether the number of evaluated points is greater than zero,
    # because we do not know how many points can be evaluated.
    assert spot_4.y.shape[0] > 0


from numpy.random import default_rng
from ..design.spacefilling import spacefilling
from ..build.kriging import Kriging
from ..utils.repair import repair_non_numeric
import numpy as np
import pandas as pd
import pylab
from scipy import optimize
from scipy.optimize import differential_evolution
from math import inf
from math import isfinite
import matplotlib.pyplot as plt
from numpy import argmin
from numpy import repeat
from numpy import sqrt
from numpy import spacing
from numpy import meshgrid
from numpy import ravel
from numpy import array
from numpy import append
from spotPython.utils.compare import selectNew
from spotPython.utils.aggregate import aggregate_mean_var
from spotPython.utils.repair import remove_nan
from spotPython.budget.ocba import get_ocba_X
import logging
import time
from spotPython.utils.progress import progress_bar


logger = logging.getLogger(__name__)
# configure the handler and formatter as needed
py_handler = logging.FileHandler(f"{__name__}.log", mode="w")
py_formatter = logging.Formatter("%(name)s %(asctime)s %(levelname)s %(message)s")
# add formatter to the handler
py_handler.setFormatter(py_formatter)
# add handler to the logger
logger.addHandler(py_handler)


class Spot:
    """
    Spot base class to handle the following tasks in a uniform manner:

    * Getting and setting parameters. This is done via the `Spot` initilaization.
    * Running surrogate based hyperparameter optimization. After the class is initialized, hyperparameter tuning
    runs can be performed via the `run` method.
    * Displaying information. The `plot` method can be used for visualizing results. The `print` methods summarizes
    information about the tuning run.

    The `Spot` class is build in a modular manner. It combines the following three components:

        1. Design
        2. Surrogate
        3. Optimizer

    For each of the three components different implementations can be selected and combined. Internal components
    are selected as default. These can be replaced by components from other packages, e.g., scikit-learn or
    scikit-optimize.

    Args:
        fun (object): objective function
        lower (numpy.array): lower bound
        upper (numpy.array): upper bound
        fun_evals (int):
            number of function evaluations
        fun_repeats (int):
            number of repeats (replicates).
        max_time (int):
            maximum time (in minutes)
        noise (bool):
            deterministic or noisy objective function
        tolerance_x (float):
            tolerance for new x solutions. Minimum distance of new solutions,
            generated by `suggest_new_X`, to already existing solutions.
            If zero (which is the default), every new solution is accepted.
        ocba_delta (int): OCBA increment (only used if `noise==True`)
        var_type (list): list of type information, can be either "num" or "factor"
        var_name (list): list of variable names, e.g., ["x1", "x2"]
        infill_criterion (string): Can be `"y"`, `"s"`, `"ei"` (negative expected improvement), or `"all"`.
        n_points (int): number of infill points
        seed (int): initial seed
        log_level (int): log level with the following settings:
            `NOTSET` (`0`),
            `DEBUG` (`10`: Detailed information, typically of interest only when diagnosing problems.),
            `INFO` (`20`: Confirmation that things are working as expected.),
            `WARNING` (`30`: An indication that something unexpected happened, or indicative of some problem in the near
                future (e.g. ‘disk space low’). The software is still working as expected.),
            `ERROR` (`40`: Due to a more serious problem, the software has not been able to perform some function.), and
            `CRITICAL` (`50`: A serious error, indicating that the program itself may be unable to continue running.)
        show_models (bool): Plot model. Currently only 1-dim functions are supported.
        show_progress (bool). Show progress bar.
        design (object): experimental design.
        design_control (dict): experimental design information stored as a dictionary with the following entries:
            "init_size": `10`, "repeats": `1`.
        surrogate (object): surrogate model. If `None`, spotPython's `kriging` is used.
        surrogate_control (dict): surrogate model information stored as a dictionary with the following entries:
            "model_optimizer": `differential_evolution`,
            "model_fun_evals": `None`,
            "min_theta": `-3.`,
            "max_theta": `3.`,
            "n_theta": `1`,
            "n_p": `1`,
            "optim_p": `False`,
            "cod_type": `"norm"`,
            "var_type": `self.var_type`,
            "use_cod_y": `False`.
        optimizer (object): optimizer. If `None`, `scipy.optimize`'s `differential_evolution` is used.
        optimizer_control (dict): information about the optimizer stored as a dictionary with the following entries:
            "max_iter": `1000`.

    Note:
        Description in the source code refers to [bart21i]:
        Bartz-Beielstein, T., and Zaefferer, M. Hyperparameter tuning approaches.
        In Hyperparameter Tuning for Machine and Deep Learning with R - A Practical Guide,
        E. Bartz, T. Bartz-Beielstein, M. Zaefferer, and O. Mersmann, Eds. Springer, 2022, ch. 4, pp. 67–114.
    """

    def __str__(self):
        return self.__class__.__name__

    def __init__(
        self,
        fun,
        lower,
        upper,
        fun_evals=15,
        fun_repeats=1,
        fun_control={},
        max_time=inf,
        noise=False,
        tolerance_x=0,
        var_type=["num"],
        var_name=None,
        all_var_name=None,
        infill_criterion="y",
        n_points=1,
        ocba_delta=0,
        seed=123,
        log_level=50,
        show_models=False,
        show_progress=False,
        design=None,
        design_control={},
        surrogate=None,
        surrogate_control={},
        optimizer=None,
        optimizer_control={},
    ):
        # small value:
        self.eps = sqrt(spacing(1))
        self.fun = fun
        self.lower = lower
        self.upper = upper
        self.var_type = var_type
        self.var_name = var_name
        self.all_var_name = all_var_name
        # Reduce dim based on lower == upper logic:
        # modifies lower, upper, and var_type
        self.to_red_dim()
        self.k = self.lower.size
        self.fun_evals = fun_evals
        self.fun_repeats = fun_repeats
        self.max_time = max_time
        self.noise = noise
        self.tolerance_x = tolerance_x
        self.ocba_delta = ocba_delta
        self.log_level = log_level
        self.show_models = show_models
        self.show_progress = show_progress
        # Random number generator:
        self.seed = seed
        self.rng = default_rng(self.seed)
        # Force numeric type as default in every dim:
        # assume all variable types are "num" if "num" is
        # specified once:
        if len(self.var_type) < self.k:
            self.var_type = self.var_type * self.k
            logger.warning("Warning: All variable types forced to 'num'.")
        self.infill_criterion = infill_criterion
        # Bounds
        de_bounds = []
        for j in range(self.k):
            de_bounds.append([self.lower[j], self.upper[j]])
        self.de_bounds = de_bounds
        # Infill points:
        self.n_points = n_points
        # Objective function related information:
        self.fun_control = {"sigma": 0, "seed": None}
        self.fun_control.update(fun_control)
        # Design related information:
        self.design = design
        if design is None:
            self.design = spacefilling(k=self.k, seed=self.seed)
        self.design_control = {"init_size": 10, "repeats": 1}
        self.design_control.update(design_control)
        # Surrogate related information:
        self.surrogate = surrogate
        self.surrogate_control = {
            "noise": self.noise,
            "model_optimizer": differential_evolution,
            "model_fun_evals": None,
            "min_theta": -3.0,
            "max_theta": 3.0,
            "n_theta": 1,
            "n_p": 1,
            "optim_p": False,
            "cod_type": "norm",
            "var_type": self.var_type,
            "seed": 124,
            "use_cod_y": False,
        }
        self.surrogate_control.update(surrogate_control)
        # If no surrogate model is specified, use the internal
        # spotPython kriging surrogate:
        if self.surrogate is None:
            # Call kriging with surrogate_control parameters:
            self.surrogate = Kriging(
                name="kriging",
                noise=self.surrogate_control["noise"],
                model_optimizer=self.surrogate_control["model_optimizer"],
                model_fun_evals=self.surrogate_control["model_fun_evals"],
                seed=self.surrogate_control["seed"],
                log_level=self.log_level,
                min_theta=self.surrogate_control["min_theta"],
                max_theta=self.surrogate_control["max_theta"],
                n_theta=self.surrogate_control["n_theta"],
                n_p=self.surrogate_control["n_p"],
                optim_p=self.surrogate_control["optim_p"],
                cod_type=self.surrogate_control["cod_type"],
                var_type=self.surrogate_control["var_type"],
                use_cod_y=self.surrogate_control["use_cod_y"],
            )
        # Optimizer related information:
        self.optimizer = optimizer
        self.optimizer_control = {"max_iter": 1000, "seed": 125}
        self.optimizer_control.update(optimizer_control)
        if self.optimizer is None:
            self.optimizer = optimize.differential_evolution
        # Logging information:
        self.counter = 0
        self.min_y = None
        self.min_X = None
        self.min_mean_X = None
        self.min_mean_y = None
        self.mean_X = None
        self.mean_y = None
        self.var_y = None
        logger.setLevel(self.log_level)
        logger.info(f"Starting the logger at level {self.log_level} for module {__name__}:")

    def to_red_dim(self):
        self.all_lower = self.lower
        self.all_upper = self.upper
        self.ident = (self.upper - self.lower) == 0
        self.lower = self.lower[~self.ident]
        self.upper = self.upper[~self.ident]
        self.red_dim = self.ident.any()
        self.all_var_type = self.var_type
        self.var_type = [x for x, y in zip(self.all_var_type, self.ident) if not y]
        if self.var_name is not None:
            self.all_var_name = self.var_name
            self.var_name = [x for x, y in zip(self.all_var_name, self.ident) if not y]

    def to_all_dim(self, X0):
        n = X0.shape[0]
        k = len(self.ident)
        X = np.zeros((n, k))
        j = 0
        for i in range(k):
            if self.ident[i]:
                X[:, i] = self.all_lower[i]
                j = j + 1
            else:
                X[:, i] = X0[:, i - j]
        return X

    def run(self):
        """
        Run spot.

        Returns:
            (object): spot
        """
        # (S-2) Initial Design:
        X0 = self.generate_design(
            size=self.design_control["init_size"],
            repeats=self.design_control["repeats"],
            lower=self.lower,
            upper=self.upper,
        )
        X0 = repair_non_numeric(X0, self.var_type)
        self.X = X0
        # (S-3): Eval initial design:
        if self.red_dim:
            X_all = self.to_all_dim(X0)
        else:
            X_all = X0
        self.y = self.fun(X=X_all, fun_control=self.fun_control)
        # TODO: Error if only nan values are returned
        logger.debug("New y value: %s", self.y)
        self.X, self.y = remove_nan(self.X, self.y)
        self.update_stats()
        # (S-4): Imputation:
        # Not implemented yet.

        self.surrogate.fit(self.X, self.y)

        # (S-5) Calling the spotLoop Function
        # and
        # (S-9) Termination Criteria, Conditions:

        timeout_start = time.time()
        while (self.counter < self.fun_evals) and (time.time() < timeout_start + self.max_time * 60):
            # OCBA (only if noise)
            if self.noise and self.ocba_delta > 0:  # and self.fun_repeats > 0 and self.design_control["repeats"] > 0:
                X_ocba = get_ocba_X(self.mean_X, self.mean_y, self.var_y, self.ocba_delta)
            else:
                X_ocba = None
            # (S-15) Compile Surrogate Results:
            X0 = self.suggest_new_X()
            X0 = repair_non_numeric(X0, self.var_type)
            # (S-16) Duplicate Handling:
            # Condition: select only X= that have min distance
            # to existing solutions
            X0, X0_ind = selectNew(A=X0, X=self.X, tolerance=self.tolerance_x)
            logger.debug("XO values are new: %s %s", X0_ind, X0)
            # 1. There are X0 that fullfil the condition.
            # Note: The number of new X0 can be smaller than self.n_points!
            if X0.shape[0] > 0:
                X0 = repeat(X0, self.fun_repeats, axis=0)
            # 2. No X0 found. Then generate self.n_points new solutions:
            else:
                self.design = spacefilling(k=self.k, seed=self.seed + self.counter)
                X0 = self.generate_design(
                    size=self.n_points, repeats=self.design_control["repeats"], lower=self.lower, upper=self.upper
                )
                X0 = repair_non_numeric(X0, self.var_type)
                logger.warning("No new XO found on surrogate. Generate new solution %s", X0)
            # (S-18): Evaluating New Solutions:
            if self.noise and self.ocba_delta > 0:
                X0 = append(X_ocba, X0, axis=0)
            if self.red_dim:
                X_all = self.to_all_dim(X0)
            else:
                X_all = X0
            y0 = self.fun(X=X_all, fun_control=self.fun_control)
            X0, y0 = remove_nan(X0, y0)
            # Append New Solutions:
            self.X = np.append(self.X, X0, axis=0)
            self.y = np.append(self.y, y0)

            # (S-10): Subset Selection for the Surrogate:
            # Not implemented yet.

            # Update stats
            self.update_stats()

            # (S-11) Surrogate Fit:
            self.surrogate.fit(self.X, self.y)

            if self.show_models:
                self.plot_model()
            # progress bar:
            if self.show_progress:
                if isfinite(self.fun_evals):
                    progress_bar(progress=self.counter / self.fun_evals)
                else:
                    progress_bar(progress=(time.time() - timeout_start) / (self.max_time * 60))
        return self

    def generate_design(self, size, repeats, lower, upper):
        return self.design.scipy_lhd(n=size, repeats=repeats, lower=lower, upper=upper)

    def update_stats(self):
        """
        Update the following stats: 1. `min_y` 2. `min_X` 3. `counter`
        If `noise` is `True`, additionally the following stats are computed: 1. `mean_X`
        2. `mean_y` 3. `min_mean_y` 4. `min_mean_X`.

        """
        self.min_y = min(self.y)
        self.min_X = self.X[argmin(self.y)]
        self.counter = self.y.size
        # Update aggregated x and y values (if noise):
        if self.noise:
            Z = aggregate_mean_var(X=self.X, y=self.y)
            self.mean_X = Z[0]
            self.mean_y = Z[1]
            self.var_y = Z[2]
            self.min_mean_y = min(self.mean_y)
            self.min_mean_X = self.mean_X[argmin(self.mean_y)]

    def suggest_new_X(self):
        """
        Compute `n_points` new infill points in natural units.
        The optimizer searches in the ranges from `lower_j` to `upper_j`.
        The method `infill()` is used as the objective function.

        Returns:
            (numpy.ndarray): `n_points` infill points in natural units, each of dim k

        Note:
            This is step (S-14a) in [bart21i].
        """
        # (S-14a) Optimization on the surrogate:
        new_X = np.zeros([self.n_points, self.k], dtype=float)

        for i in range(self.n_points):
            if self.optimizer.__name__ == "dual_annealing":
                result = self.optimizer(func=self.infill, bounds=self.de_bounds)
            elif self.optimizer.__name__ == "differential_evolution":
                result = self.optimizer(
                    func=self.infill,
                    bounds=self.de_bounds,
                    maxiter=self.optimizer_control["max_iter"],
                    seed=self.optimizer_control["seed"],
                    # popsize=10,
                    # updating="deferred"
                )
            elif self.optimizer.__name__ == "direct":
                result = self.optimizer(func=self.infill, bounds=self.de_bounds, eps=1e-2)
            elif self.optimizer.__name__ == "shgo":
                result = self.optimizer(func=self.infill, bounds=self.de_bounds)
            elif self.optimizer.__name__ == "basinhopping":
                result = self.optimizer(func=self.infill, x0=self.min_X)
            else:
                result = self.optimizer(func=self.infill, bounds=self.de_bounds)
            new_X[i][:] = result.x
        return new_X

    def infill(self, x):
        """
        Infill (acquisition) function. Evaluates one point on the surrogate via `surrogate.predict(x.reshape(1,-1))`,
        if `sklearn` surrogates are used or `surrogate.predict(x.reshape(1,-1), return_val=self.infill_criterion)`
        if the internal surrogate `kriging` is selected.
        This method is passed to the optimizer in `suggest_new_X`, i.e., the optimizer is called via
        `self.optimizer(func=self.infill)`.

        Args:
            x (array): point in natural units with shape `(1, dim)`.

        Returns:
            (numpy.ndarray): value based on infill criterion, e.g., `"ei"`. Shape `(1,)`.
                The objective function value `y` that is used as a base value for the
                infill criterion is calculated in natural units.

        Note:
            This is step (S-12) in [bart21i].
        """
        # (S-12) Objective Function on the Surrogate (Predict)
        # res = self.surrogate.predict(x.reshape(1, -1))
        # if isinstance(self.surrogate, Kriging):
        #     if self.infill_criterion == "ei":
        #         y = -1.0 * res[2]  # ei
        #     else:
        #         y = res[0]  # f
        # else:
        #     y = res  # sklearn etc.
        # return y
        if isinstance(self.surrogate, Kriging):
            return self.surrogate.predict(x.reshape(1, -1), return_val=self.infill_criterion)
        else:
            return self.surrogate.predict(x.reshape(1, -1))

    def plot_progress(self, show=True, log_y=False):
        """
        Generate progress plot, i.e., plot y versus x values.
        Usually called after the run is finished.

        Args:
            show (bool, optional): Show plot. Defaults to True.
            log_y (bool, optional): log y-axis. Defaults to False.
        """
        fig = pylab.figure(figsize=(9, 6))
        # TODO: Consider y_min of the initial design, e.g., via:
        # best_y = list(itertools.repeat(min(y), self.init_size))
        s_y = pd.Series(self.y)
        s_c = s_y.cummin()

        ax = fig.add_subplot(211)
        ax.plot(range(len(s_c)), s_c)
        if log_y:
            ax.set_yscale("log")
        if show:
            pylab.show()

    def plot_model(self, y_min=None, y_max=None):
        """
        Plot the model fit for 1-dim objective functions.

        Args:
            y_min (float, optional): y range, lower bound.
            y_max (float, optional): y range, upper bound.
        """
        if self.k == 1:
            X_test = np.linspace(self.lower[0], self.upper[0], 100)
            y_test = self.fun(X=X_test.reshape(-1, 1), fun_control=self.fun_control)
            if isinstance(self.surrogate, Kriging):
                y_hat = self.surrogate.predict(X_test[:, np.newaxis], return_val="y")
            else:
                y_hat = self.surrogate.predict(X_test[:, np.newaxis])
            plt.plot(X_test, y_hat, label="Model")
            plt.plot(X_test, y_test, label="True function")
            plt.scatter(self.X, self.y, edgecolor="b", s=20, label="Samples")
            plt.scatter(self.X[-1], self.y[-1], edgecolor="r", s=30, label="Last Sample")
            if self.noise:
                plt.scatter(self.min_mean_X, self.min_mean_y, edgecolor="g", s=30, label="Best Sample (mean)")
            else:
                plt.scatter(self.min_X, self.min_y, edgecolor="g", s=30, label="Best Sample")
            plt.xlabel("x")
            plt.ylabel("y")
            plt.xlim((self.lower[0], self.upper[0]))
            if y_min is None:
                y_min = min(min(self.y), min(y_test))
            if y_max is None:
                y_max = max(max(self.y), max(y_test))
            plt.ylim((y_min, y_max))
            plt.legend(loc="best")
            # plt.title(self.surrogate.__class__.__name__ + ". " + str(self.counter) + ": " + str(self.min_y))
            if self.noise:
                plt.title(
                    str(self.counter)
                    + ". y (noise): "
                    + str(np.round(self.min_y, 6))
                    + " y mean: "
                    + str(np.round(self.min_mean_y, 6))
                )
            else:
                plt.title(str(self.counter) + ". y: " + str(np.round(self.min_y, 6)))
            plt.show()

    def print_results(self):
        """
        Print results from the run:
            1. min y
            2. min X
            If `noise == True`, additionally the following values are printed:
            3. min mean y
            4. min mean X
        """
        print(f"min y: {self.min_y}")
        res = self.to_all_dim(self.min_X.reshape(1, -1))
        for i in range(res.shape[1]):
            if self.all_var_name is None:
                print("x" + str(i) + ":", res[0][i])
            else:
                print(self.all_var_name[i] + ":", res[0][i])
        if self.noise:
            res = self.to_all_dim(self.min_mean_X.reshape(1, -1))
            print(f"min mean y: {self.min_mean_y}")
            for i in range(res.shape[1]):
                if self.all_var_name is None:
                    print("x" + str(i) + ":", res[0][i])
                else:
                    print(self.all_var_name[i] + ":", res[0][i])

    def chg(self, x, y, z0, i, j):
        z0[i] = x
        z0[j] = y
        return z0

    def plot_contour(self, i=0, j=1, min_z=None, max_z=None, show=True):
        """
        This function plots surrogates of any dimension.
        Args:
            show (boolean):
                If `True`, the plots are displayed.
                If `False`, `plt.show()` should be called outside this function.
        """
        fig = pylab.figure(figsize=(9, 6))
        n_grid = 100
        # lower and upper
        x = np.linspace(self.lower[i], self.upper[i], num=n_grid)
        y = np.linspace(self.lower[j], self.upper[j], num=n_grid)
        X, Y = meshgrid(x, y)
        # Predict based on the optimized results
        z0 = np.mean(np.array([self.lower, self.upper]), axis=0)
        zz = array([self.surrogate.predict(array([self.chg(x, y, z0, i, j)])) for x, y in zip(ravel(X), ravel(Y))])
        zs = zz[:, 0]
        Z = zs.reshape(X.shape)
        if min_z is None:
            min_z = np.min(Z)
        if max_z is None:
            max_z = np.max(Z)
        contour_levels = 30
        ax = fig.add_subplot(221)
        # plot predicted values:
        plt.contourf(X, Y, Z, contour_levels, zorder=1, cmap="jet", vmin=min_z, vmax=max_z)
        if self.var_name is None:
            plt.xlabel("x" + str(i))
            plt.ylabel("x" + str(j))
        else:
            plt.xlabel("x" + str(i) + ": " + self.var_name[i])
            plt.ylabel("x" + str(j) + ": " + self.var_name[j])
        plt.title("Surrogate")
        pylab.colorbar()
        #
        ax = fig.add_subplot(222, projection="3d")
        ax.plot_surface(X, Y, Z, rstride=3, cstride=3, alpha=0.9, cmap="jet", vmin=min_z, vmax=max_z)
        if self.var_name is None:
            plt.xlabel("x" + str(i))
            plt.ylabel("x" + str(j))
        else:
            plt.xlabel("x" + str(i) + ": " + self.var_name[i])
            plt.ylabel("x" + str(j) + ": " + self.var_name[j])
        #
        pylab.show()

    def print_importance(self):
        if self.surrogate.n_theta > 1:
            theta = np.power(10, self.surrogate.theta)
            print("Importance relative to the most important parameter:")
            imp = 100 * theta / np.max(theta)
            if self.var_name is None:
                for i in range(len(imp)):
                    print("x", i, ": ", imp[i])
            else:
                for i in range(len(imp)):
                    print(self.var_name[i] + ": ", imp[i])
        else:
            print("Importantance requires more than one theta values (n_theta>1).")



def test_set_de_bounds():
    """
    Test set_de_bounds
    """
    import numpy as np
    from spotPython.spot import spot
    from spotPython.build.kriging import Kriging

    X_train = np.array([[1., 2.],
                        [2., 4.],
                        [3., 6.]])
    y_train = np.array([1., 2., 3.])

    # Without Noise:

    S = Kriging(name='kriging',
                seed=123,
                log_level=50,
                n_theta=1,
                noise=False,
                cod_type="norm")
    S.fit(X_train, y_train)
    res = [[S.min_theta, S.max_theta]]
    S.set_de_bounds()
    assert np.array_equal(S.de_bounds, res)

    # Return:

    # assume 1 theta, fixed p, no noise:

    new_theta_p_Lambda = np.array([-2.11625348])
    S.extract_from_bounds(new_theta_p_Lambda)

    res_theta = np.array([-2.11625348])
    assert np.array_equal(S.theta, res_theta)
    # p should be 2.0
    res_p = np.array([2.0])
    assert np.array_equal(S.p, res_p)
    # Lambda should be None
    assert S.Lambda is None


    # With Noise:

    S = Kriging(name='kriging',  seed=123, log_level=50, n_theta=1, noise=True, cod_type="norm")
    S.fit(X_train, y_train)

    res = [[S.min_theta, S.max_theta], [S.min_Lambda, S.max_Lambda]]

    S.set_de_bounds()

    assert np.array_equal(S.de_bounds, res)

    # Return:

    # assume 1 theta, fixed p, noise:

    new_theta_p_Lambda = np.array([-2.11625348, 0.1234])
    S.extract_from_bounds(new_theta_p_Lambda)

    res_theta = np.array([-2.11625348])
    assert np.array_equal(S.theta, res_theta)
    # p should be 2.0
    res_p = np.array([2.0])
    assert np.array_equal(S.p, res_p)
    # Lambda should be 0.1234
    res_Lambda = 0.1234
    assert S.Lambda == res_Lambda

    # 2nd Test Series
    from spotPython.build.kriging import Kriging
    import numpy as np
    import copy
    import matplotlib.pyplot as plt
    from numpy import append, ndarray, multiply, isinf, linspace, meshgrid, ravel
    from numpy import ones, zeros, arange, log, var, float64
    from numpy import spacing, empty_like
    from numpy import array
    from spotPython.design.spacefilling import spacefilling

    # One-dim objective function
    ni = 11
    nat_X = linspace(start=0, stop=10, num=ni).reshape(-1, 1)
    nat_y = np.squeeze(nat_X +1)
    S = Kriging(name='kriging',
                min_theta = -3,
                max_theta = 2,
                seed=124)
    S.nat_X = copy.deepcopy(nat_X)
    S.nat_y = copy.deepcopy(nat_y)
    S.n = S.nat_X.shape[0]
    S.k = S.nat_X.shape[1]
    S.cod_X = empty_like(S.nat_X)
    S.cod_y = empty_like(S.nat_y)
    # assume all variable types are "num" if "num" is
    # specified once:
    if len(S.var_type) == 1:
        S.var_type = S.var_type * S.k
    S.num_mask = array(list(map(lambda x: x == "num", S.var_type)))
    S.factor_mask = array(list(map(lambda x: x == "factor", S.var_type)))

    S.nat_to_cod_init()
    S.theta = zeros(S.n_theta)
    # TODO: Currently not used:
    S.x0_theta = ones((S.n_theta,)) * S.n / (100 * S.k)
    S.p = ones(S.n_p) * 2.0

    S.pen_val = S.n * log(var(S.nat_y)) + 1e4
    S.negLnLike = None

    S.gen = spacefilling(k=S.k, seed=S.seed)

    # matrix related
    S.LnDetPsi = None
    S.Psi = zeros((S.n, S.n), dtype=float64)
    S.psi = zeros((S.n, 1))
    S.one = ones(S.n)
    S.mu = None
    S.U = None
    S.SigmaSqr = None
    S.Lambda = None
    # build_Psi() and build_U() are called in fun_likelihood
    S.set_de_bounds()
    # 1. check default (only theta):
    assert S.de_bounds[0][0] == S.min_theta
    assert S.de_bounds[0][1] == S.max_theta
    # 2. Check theta and p:
    S = Kriging(name='kriging',
                min_theta = -4,
                max_theta = 5,
                optim_p=True,
                seed=124)

    S.set_de_bounds()
    assert S.de_bounds[0][0] == S.min_theta
    assert S.de_bounds[0][1] == S.max_theta
    assert S.de_bounds[1][0] == S.min_p
    assert S.de_bounds[1][1] == S.max_p

    # 3. Check theta, p, and Lambda:
    S = Kriging(name='kriging',
                min_theta = -4,
                max_theta = 5,
                optim_p=True,
                noise=True,
                seed=124)

    S.set_de_bounds()
    assert S.de_bounds[0][0] == S.min_theta
    assert S.de_bounds[0][1] == S.max_theta
    assert S.de_bounds[1][0] == S.min_p
    assert S.de_bounds[1][1] == S.max_p
    assert S.de_bounds[2][0] == S.min_Lambda
    assert S.de_bounds[2][1] == S.max_Lambda

    # 3. Check theta and Lambda:
    S = Kriging(name='kriging',
                min_theta = -4,
                max_theta = 5,
                optim_p=False,
                noise=True,
                seed=124)

    S.set_de_bounds()
    assert S.de_bounds[0][0] == S.min_theta
    assert S.de_bounds[0][1] == S.max_theta
    assert S.de_bounds[1][0] == S.min_Lambda
    assert S.de_bounds[1][1] == S.max_Lambda



def scale(X, lower, upper):
    """
    Sample scaling from unit hypercube to different bounds.

    Converts a sample from `[0, 1)` to `[a, b)`.
    Note: equal lower and upper bounds are feasible.
    The following transformation is used:

    `(b - a) * X + a`

    Args:
    X (array):
        Sample to scale.
    lower (array):
        lower bound of transformed data.
    upper (array):
        upper bounds of transformed data.

    Returns:
    (array):
        Scaled sample.

    Examples:
    Transform three samples in the unit hypercube to (lower, upper) bounds:

    >>> import numpy as np
    >>> from scipy.stats import qmc
    >>> from spotPython.utils.transform import scale
    >>> lower = np.array([6, 0])
    >>> upper = np.array([6, 5])
    >>> sample = np.array([[0.5 , 0.75],
    >>>             [0.5 , 0.5],
    >>>             [0.75, 0.25]])
    >>> scale(sample, lower, upper)

    """
    # Checking that X is within (0,1) interval
    if (X.max() > 1.0) or (X.min() < 0.0):
        raise ValueError("Sample is not in unit hypercube")

    for i in range(X.shape[1]):
        if lower[i] == upper[i]:
            X[:, i] = lower[i]
        else:
            X[:, i] = X[:, i] * (upper[i] - lower[i]) + lower[i]
    return X



def test_nat_to_cod():
    """
    Test nat_to_cod and cod_to_nat
    """
from spotPython.fun.objectivefunctions import analytical
from spotPython.design.factorial import factorial
from spotPython.build.kriging import Kriging
import numpy as np

gen = factorial(3)
rng = np.random.RandomState(1)
lower = np.array([-1,-1])
upper = np.array([0,0])
fun = analytical().fun_linear
X = gen.full_factorial(3)
X = 10*X
y = fun(X)

S = Kriging(name='kriging',  seed=123)
S.fit(X,y)
X2 = 2*X

Y = np.empty_like(X2)
T = np.empty_like(X2)
for i in range(S.n):
    T[i] = S.nat_to_cod_x(X2[i])
    Y[i] = S.cod_to_nat_x(T[i])

assert np.array_equal(Y, X2)



def test_generate_design():
    """
    Test generate_design method
    """
    import numpy as np
    from spotPython.fun.objectivefunctions import analytical
    from spotPython.spot.spot import Spot
    fun = analytical().fun_branin_factor

    ni = 7

    spot_test = Spot(
        fun=fun,
        lower=np.array([-5,-0,1]),
        upper=np.array([10,15,3]),
        seed=1,
        design_control={"init_size": ni}
    )

    X = spot_test.generate_design(size=spot_test.design_control["init_size"],
                                  repeats=spot_test.design_control["repeats"],
                                  lower=spot_test.lower, 
                                  upper=spot_test.upper)
    for i in range(X.shape[1]):
        for j in range(X.shape[0]):
            assert X[j,i] >= spot_test.lower[i]
            assert X[j,i] <= spot_test.upper[i]
    assert X.shape[0] == ni
    assert X.shape[1] == 3

    X2 = spot_test.generate_design(size=spot_test.design_control["init_size"],
                                   repeats=spot_test.design_control["repeats"],
                                   lower=spot_test.lower, 
                                   upper=spot_test.upper)

    spot_test = Spot(
        fun=fun,
        lower=np.array([-5,-0,1]),
        upper=np.array([10,15,3]),
        seed=1,
        design_control={"init_size": ni}
    )

    X3 = spot_test.generate_design(size=spot_test.design_control["init_size"],
                                   repeats=spot_test.design_control["repeats"],
                                   lower=spot_test.lower, 
                                   upper=spot_test.upper)

    spot_test = Spot(
        fun=fun,
        lower=np.array([-5,-0,1]),
        upper=np.array([10,15,3]),
        seed=2,
        design_control={"init_size": ni}
    )

    X4 = spot_test.generate_design(size=spot_test.design_control["init_size"],
                                   repeats=spot_test.design_control["repeats"],
                                   lower=spot_test.lower, 
                                   upper=spot_test.upper)

    assert (X != X2).any()
    assert (X == X3).any()
    assert (X != X4).any()
    assert (X2 != X3).any()
    assert (X2 != X4).any()
    assert (X3 != X4).any()



import pandas as pd
import numpy as np


def aggregate_mean_var(X, y, sort=False):
    """
    Aggregate array to mean.

    Args:
        X (numpy.ndarray): X array, shape `(n, k)`.
        y (numpy.ndarray): values, shape `(n,)`.

    Returns:
        (numpy.ndarray): aggregated `X` values, shape `(n-m, k)`, if `m`duplicates in `X`.
        (numpy.ndarray): aggregated (mean per group) `y` values, shape `(1,)`, if `m`duplicates in `X`.
        (numpy.ndarray): aggregated (variance per group) `y` values, shape `(1,)`, if `m`duplicates in `X`.
    """
    df = pd.DataFrame(X, dtype=pd.Float64Dtype)
    # df.columns=["X"+str(i) for i in range(df.shape[1])]
    df = df.assign(y=y)
    df_m = df.groupby(list(df.columns.difference(["y"])), as_index=False, sort=sort).mean()
    df_var = df.groupby(list(df.columns.difference(["y"])), as_index=False, sort=sort).var()
    A = df_m.to_numpy(dtype=np.float64)
    B = df_var.to_numpy()
    return np.delete(A, -1, 1), A[:, -1], B[:, -1]


def get_ranks(x):
    """
    Returns a numpy array containing ranks of numbers within an input numpy array x:

    Examples:

    get_ranks([2, 1])
    [1, 0]

    get_ranks([20, 10, 100])
    [1, 0, 2]

    Args:
        x (numpy.ndarray): numpy array

    Returns:
        (numpy.ndarray): ranks

    """
    ts = x.argsort()
    ranks = np.empty_like(ts)
    ranks[ts] = np.arange(len(x))
    return ranks
